#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Bridge between OSC and MQTT."""

from __future__ import absolute_import, unicode_literals

import argparse
import logging
import shlex
import sys
import time

from collections import OrderedDict

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

import liblo
import paho.mqtt.client as mqtt

from .converter import Osc2MqttConverter, ConversionRule
from .util import as_bool, parse_hostport, parse_list


log = logging.getLogger('osc2mqtt')


def read_config(fn, options="options"):
    config = {'rules': OrderedDict()}
    defaults = dict(
        match = '^/?(.*)',
        address = r'/\1',
        topic = r'\1',
        type = 'struct',
        format = 'B',
        from_mqtt = None,
        from_osc = None,
        osctags = None
    )

    if fn:
        cp = configparser.RawConfigParser(defaults)
        cp.read(fn)
        if cp.has_section(options):
            config.update(i for i in cp.items(options)
                          if i not in cp.items('DEFAULT'))

        for section in cp.sections():
            if section.startswith(':'):
                name = section[1:]
                config['rules'][name] = dict(cp.items(section))

    subscriptions = parse_list(config.get('subscriptions', '#'))
    config['subscriptions'] = []
    encode = ((lambda s: s.encode('utf-8'))
              if isinstance(b'', str) else (lambda s: s))
    for sub in subscriptions:
        config['subscriptions'].append((encode(sub), 0))

    return config


class Osc2MqttBridge(object):
    def __init__(self, config, converter):
        """Setup OSC and MQTT servers.

        @param config: configuration directory
        @param converter: Osc2MqttConverter instance

        """
        self.converter = converter
        self.config = config
        self.mqtt_host, self.mqtt_port = parse_hostport(
            config.get("mqtt_broker", "localhost"), 1883)
        self.osc_port = int(config.get("osc_port", 9001))
        self.osc_receiver = config.get("osc_receiver")
        self.subscriptions = config.get("subscriptions", ['#'])

        if self.osc_receiver:
            host, port = parse_hostport(self.osc_receiver, 9000)
            self.osc_receiver = liblo.Address(host, port, liblo.UDP)

        self.mqttclient = mqtt.Client(config.get("client_id", "osc2mqtt"))
        self.mqttclient.on_connect = self.mqtt_connect
        self.mqttclient.on_disconnect = self.mqtt_disconnect
        self.mqttclient.on_message = self.handle_mqtt

        self.oscserver = liblo.ServerThread(self.osc_port)
        self.oscserver.add_method(None, None, self.handle_osc)

    def start(self):
        """Start MQTT client and OSC listener."""
        log.info("Connecting to MQTT broker %s:%s ...",
            self.mqtt_host, self.mqtt_port)
        self.mqttclient.connect(self.mqtt_host, self.mqtt_port)

        log.debug("Starting MQTT thread...")
        self.mqttclient.loop_start()

        log.info("Starting OSC server listening on port %s ...", self.osc_port)
        self.oscserver.start()

    def stop(self):
        """Method docstring."""
        log.info("Stopping OSC server ...")
        self.oscserver.stop()
        log.debug("Stopping MQTT thread ...")
        self.mqttclient.loop_stop()
        log.info("Disconnecting from MQTT broker ...")
        self.mqttclient.disconnect()

    def mqtt_connect(self, client, userdata, flags, rc):
        log.debug("MQTT connect: %s", mqtt.connack_string(rc))
        if rc == 0:
            client.subscribe(self.subscriptions)

    def mqtt_disconnect(self, client, userdata, rc):
        log.debug("MQTT disconnect: %s", mqtt.error_string(rc))

    def handle_mqtt(self, client, userdata, msg):
        log.debug("MQTT recv: %s %r", msg.topic, msg.payload)
        res = self.converter.from_mqtt(msg.topic, msg.payload)
        if res:
            if self.osc_receiver:
                log.debug("OSC send: %s %r", *res)
                self.oscserver.send(self.osc_receiver, res[0], *res[1])
        else:
            log.debug("No rule match for MQTT topic '%s'.", msg.topic)

    def handle_osc(self, oscaddr, values, tags, clientaddr, userdata):
        log.debug("OSC recv: %s %r", oscaddr, values)
        res = self.converter.from_osc(oscaddr, values, tags)
        if res:
            log.debug("MQTT publish: %s %r", *res)
            self.mqttclient.publish(*res)
        else:
            log.debug("No rule match for OSC address '%s'.", oscaddr)


def main(args=None):
    ap = argparse.ArgumentParser(description=__doc__.splitlines()[0])
    ap.add_argument(
        '-c', '--config',
        metavar='FILENAME',
        default='osc2mqtt.ini',
        help="Read configuration from given filename")
    ap.add_argument(
        '-p', '--osc-port',
        type=int,
        metavar='PORT',
        default=9001,
        help="Local OSC server (UDP) port (default: %(default)s)")
    ap.add_argument(
        '-m', '--mqtt-broker',
        metavar='ADDR[:PORT]',
        default='localhost:1883',
        help="MQTT broker addr[:port] (default: %(default)s)")
    ap.add_argument(
        '-o', '--osc-receiver',
        metavar='ADDR[:PORT]',
        help='Also bridge MQTT to OSC receiver addr[:port] via UDP '
             '(default: one-way)')
    ap.add_argument(
        '-v', '--verbose',
        action="store_true",
        help='Enable verbose logging')

    args = ap.parse_args(args if args is not None else sys.argv[1:])

    cfg = read_config(args.config)

    for opt in ('mqtt_broker', 'osc_port', 'osc_receiver', 'verbose'):
        argval = getattr(args, opt)
        if opt not in cfg or argval != ap.get_default(opt):
            cfg[opt] = argval

    logging.basicConfig(level=logging.DEBUG
        if as_bool(cfg["verbose"]) else logging.INFO)

    converter = Osc2MqttConverter(cfg["rules"])
    osc2mqtt = Osc2MqttBridge(cfg, converter)
    osc2mqtt.start()

    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        log.info("Interrupted.")
    finally:
        osc2mqtt.stop()
        log.info("Done.")


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]) or 0)



# -*- coding: utf-8 -*-
"""Convert between MQTT message payload data and OSC arguments types."""

from __future__ import absolute_import, unicode_literals

import array
import json
import re
import struct
import logging

from collections import namedtuple

from .util import as_bool, parse_list

try:
    # Python 3.2+
    from functools import lru_cache
except ImportError:
    try:
        # try backport available on PyPI
        from backports.functools_lru_cache import lru_cache
    except ImportError:
        # finally, backport included with package
        from .lru_cache import lru_cache


log = logging.getLogger(__name__)


class ConfigError(Exception):
    """Raised when configuration file can not be parse correctly."""
    pass


ConversionRule = namedtuple(
    'ConversionRule',
    [
        'match',
        'address',
        'topic',
        'type',
        'format',
        'from_mqtt',
        'from_osc',
        'osctags'
    ])


class Osc2MqttConverter(object):
    """Convert MQTT topic and payload to OSC address and values and vice-versa.
    """

    _converters = {
        'f': float,
        'float': float,
        'i': int,
        'int': int,
        's': str,
        'str': str,
        'b': as_bool,
        'bool': as_bool
    }

    def __init__(self, rules):
        """Initialize converter with a set of conversion rules.

        @param rules: a dict of ConversionRules instances, keyed by name

        """
        self.rules = {}

        for name, rule in rules.items():
            try:
                if rule["from_mqtt"]:
                    rule["from_mqtt"] = [self._converters.get(f)
                        for f in parse_list(rule["from_mqtt"])]

                if rule["from_osc"]:
                    rule["from_osc"] = [self._converters.get(f)
                        for f in parse_list(rule["from_osc"])]

                self.rules[name] = ConversionRule(**rule)
            except Exception as exc:
                raise ConfigError("Malformed conversion rule: %s" % exc)

    @lru_cache()
    def match_rule(self, topicoraddr):
        """Match MQTT topic or OSC address against a rule regex.

        @param topicoraddr: MQTT topic or OSC address pattern string
        @return ConversionRule: the conversion rule instance, which matched
            topicoraddr, or None, if no match was found

        """
        for name, rule in self.rules.items():
            m = re.search(rule.match, topicoraddr)
            if m:
                log.debug("Rule '%s' match on: %s", name, topicoraddr)
                return rule

    def from_mqtt(self, topic, payload):
        """Convert MQTT message to OSC.

        The MQTT message payload (an opaque byte string) can be encoded in
        several forms, for example:

        1. A JSON, msgpack, etc. string.
        2. An ASCII string representation of an integer or float
        3. An integer directly encoded as the byte value
           a. signed or
           b. unsigned
        4. A multi-byte value, like a word, long, float, double, etc. in signed,
           unsigned, big and little endian varieties
        5. An array (i.e. sequence) of values encoded as 3) or 4)

        1) and 2) can be decoded with JSON.loads(), 3a) can be passed to ord() and
        3a/b) and 4) can be decoded with struct.unpack(), where one needs to know
        the type, the endianess and signedness. 5) can be decoded, for example,
        with the array module or bytearray(), if the type of the values is uint8.

        """
        rule = self.match_rule(topic)

        if rule:
            addr = re.sub(rule.match, rule.address, topic)
            values = self.decode_values(payload, rule)
            log.debug("Decoded payload to values: %r --> %r", payload, values)

            if rule.osctags:
                values = tuple(zip(rule.osctags, values))

            return addr, values

    def decode_values(self, data, rule):
        """Decode MQTT message payload byte string into Python values."""
        if rule.type == 'json':
            values = json.loads(data.decode(rule.format or 'utf-8'))
        elif rule.type == 'struct':
            values = struct.unpack_from(rule.format, data)
        elif rule.type == 'array':
            values = tuple(array.array(rule.format, data))
        elif rule.type == 'string':
            values = (data.decode(rule.format or 'utf-8'),)
        else:
            values = (data,)

        if rule.from_mqtt:
            values = tuple(func(value) if func else value
                           for func, value in zip(rule.from_mqtt, values))

        return values

    def from_osc(self, addr, values, tags):
        """Convert OSC message to MQTT.

        Since OSC messages always specify the types of their values, only the
        'type' and 'format' of the matching conversion rule is used to encode
        the OSC values into an MQTT message payload string.

        """
        rule = self.match_rule(addr)

        if rule:
            topic = re.sub(rule.match, rule.topic, addr)
            data = self.encode_values(values, rule)
            log.debug("Encoded values to payload: %r --> %r", values, data)
            return topic, data

    def encode_values(self, values, rule):
        """Encode Python values into MQTT message payload."""
        if rule.from_osc:
            values = tuple(func(value) if func else value
                           for func, value in zip(rule.from_osc, values))

        if rule.type == 'json':
            return json.dumps(values)
        elif rule.type == 'struct':
            return bytearray(struct.pack(rule.format, *values))
        elif type == 'array':
            return bytearray(array.array(rule.format, values).tostring())
        elif rule.type == 'string':
            return "".join(str(s) for s in values)
        else:
            if len(values) == 1:
                return str(values[0]).encode()
            else:
                return str(values).encode()



# -*- coding: utf-8 -*-
"""Utility functions for reading osc2mqtt configuration."""

from __future__ import absolute_import, unicode_literals

import shlex


def as_bool(val):
    return str(val).lower() in ('1', 'enabled', 'on', 't', 'true', 'y', 'yes')


def parse_hostport(addr, port=9000):
    if ('::' in addr and ']:' in addr) or ('::' not in addr and ':' in addr):
        host, port = addr.rsplit(':', 1)
    else:
        host = addr

    if host.startswith('[') and host.endswith(']'):
        host = host[1:-1]

    return (host, int(port))


def parse_list(s):
    lexer = shlex.shlex(s, posix=True)
    lexer.whitespace = ','
    lexer.whitespace_split = True
    lexer.commenters = ''
    return [token.strip() for token in lexer]
